/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamah264enc.h"

#include <gst/pbutils/pbutils.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_h264_enc_debug);
#define GST_CAT_DEFAULT gst_ama_h264_enc_debug

typedef enum
{
  GST_AMA_H264_PROFILE_BASELINE = 0,
  GST_AMA_H264_PROFILE_MAIN,
  GST_AMA_H264_PROFILE_HIGH,
  GST_AMA_H264_PROFILE_HIGH10,
  GST_AMA_H264_PROFILE_HIGH10_INTRA,
} GstAMAH264Profile;

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS
    ("video/x-h264, "
        "stream-format=(string)byte-stream, "
        "alignment=(string)au, "
        "profile= (string) { baseline, main, high, high-10, high-10-intra }, "
        "level= (string) { 1, 1.1, 1.2, 1.3, 2, 2.1, 2.2, 3, 3.1, 3.2, "
            "4, 4.1, 4.2, 5, 5.1, 5.2 };"));

G_DEFINE_TYPE (GstAMAH264Enc, gst_ama_h264_enc, GST_TYPE_AMA_ENC);
GST_ELEMENT_REGISTER_DEFINE (ama_h264enc, "ama_h264enc", GST_RANK_NONE, GST_TYPE_AMA_H264_ENC);

static guint32
gst_ama_h264_enc_profile_from_string (const gchar * str)
{
  if (!g_strcmp0 (str, "baseline")) {
    return GST_AMA_H264_PROFILE_BASELINE;
  } else if (!g_strcmp0 (str, "main")) {
    return GST_AMA_H264_PROFILE_MAIN;
  } else if (!g_strcmp0 (str, "high")) {
    return GST_AMA_H264_PROFILE_HIGH;
  } else if (!g_strcmp0 (str, "high-10")) {
    return GST_AMA_H264_PROFILE_HIGH10;
  } else if (!g_strcmp0 (str, "high-10-intra")) {
    return GST_AMA_H264_PROFILE_HIGH10_INTRA;
  }

  g_assert_not_reached ();
}

static void
gst_ama_h264_enc_configure_encoder (GstAMAEnc * self, GstCaps * outcaps,
    XmaEncoderProperties * enc_props)
{
  GstStructure *s = gst_caps_get_structure (outcaps, 0);
  const gchar *profile_str = gst_structure_get_string (s, "profile");
  const gchar *level_str = gst_structure_get_string (s, "level");

  enc_props->hwencoder_type = XMA_H264_ENCODER_TYPE;

  if (profile_str) {
    enc_props->profile = gst_ama_h264_enc_profile_from_string (profile_str);
  }

  if (level_str) {
    enc_props->level = gst_codec_utils_h264_get_level_idc (level_str);
  }
}

static void
gst_ama_h264_enc_init (GstAMAH264Enc * enc)
{
}

static void
gst_ama_h264_enc_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  if (prop_id >= GST_AMA_ENC_RANGED_PROPERTY_FIRST) {
    G_OBJECT_CLASS (gst_ama_h264_enc_parent_class)->set_property (object,
        prop_id, value, pspec);
    return;
  }

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_h264_enc_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  if (prop_id >= GST_AMA_ENC_RANGED_PROPERTY_FIRST) {
    G_OBJECT_CLASS (gst_ama_h264_enc_parent_class)->get_property (object,
        prop_id, value, pspec);
    return;
  }

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_h264_enc_class_init (GstAMAH264EncClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstAMAEncClass *amaenc_class = GST_AMA_ENC_CLASS (klass);

  gobject_class->set_property = gst_ama_h264_enc_set_property;
  gobject_class->get_property = gst_ama_h264_enc_get_property;

  amaenc_class->configure_encoder = gst_ama_h264_enc_configure_encoder;

  amaenc_class->qp_max = 51;
  amaenc_class->min_qp_max = XMA_ENC_MIN_QP_MAX_VALUE;
  amaenc_class->max_qp_max = XMA_ENC_MAX_QP_MAX_VALUE;
  amaenc_class->max_qp_default = XMA_ENC_MAX_QP_DEFAULT;
  amaenc_class->b_frames_max = XMA_ENC_MAX_BF_VALUE;
  gst_ama_enc_class_init_ranged_properties (amaenc_class);

  gst_element_class_add_static_pad_template (element_class, &src_template);

  gst_element_class_set_static_metadata (element_class,
      "Xilinx AMA H264 encoder", "Encoder/Video",
      "Xilinx H264 Encoder", "Xilinx Inc., https://www.xilinx.com");

  GST_DEBUG_CATEGORY_INIT (gst_ama_h264_enc_debug, "ama_h264enc", 0,
      "Xilinx AMA H264 encoder");
}
