/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamascaler.h"
#include "gstamacommon.h"
#include "gstxmaframeallocator.h"
#include "gstxrm.h"

#include <stdio.h>

#include <xmaerror.h>
#include <xmaparam.h>
#include <xmascaler.h>

#include <gst/base/gstbasetransform.h>
#include <gst/video/video-info.h>
#include <gst/video/video-format.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_scaler_debug);
#define GST_CAT_DEFAULT gst_ama_scaler_debug

enum
{
  PROP_0,
  PROP_DEVICE,
  PROP_ENABLE_PIPELINE,
  PROP_CROP,
  PROP_CROP_TOP,
  PROP_CROP_LEFT,
  PROP_CROP_WIDTH,
  PROP_CROP_HEIGHT,
};

#define ABRSCALER_VIDEO_FORMAT \
    ("video/x-raw(" GST_CAPS_FEATURE_MEMORY_AMA_MEMORY "), " \
        "format = (string) " GST_AMA_ALL_VIDEO_FORMATS "," \
        "width = (int) [ " G_STRINGIFY (GST_AMA_MIN_HEIGHT) ", 3840 ], " \
        "height = (int) [ " G_STRINGIFY (GST_AMA_MIN_HEIGHT) ", 3840 ], " \
        "framerate = (fraction) [ 0/1, MAX ]")

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (ABRSCALER_VIDEO_FORMAT)
    );

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src_%u",
    GST_PAD_SRC,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS (ABRSCALER_VIDEO_FORMAT)
    );

struct _GstAMAScalerPad
{
  GstPad parent;

  guint index;
  GstVideoInfo vinfo;
  gboolean half_framerate;
};

#define GST_TYPE_AMA_ABRSCALER_PAD (gst_ama_scaler_pad_get_type ())
G_DECLARE_FINAL_TYPE (GstAMAScalerPad, gst_ama_scaler_pad, GST, AMA_SCALER_PAD, GstPad)
G_DEFINE_TYPE (GstAMAScalerPad, gst_ama_scaler_pad, GST_TYPE_PAD);

static void
gst_ama_scaler_pad_init (GstAMAScalerPad * self)
{
}

static void
gst_ama_scaler_pad_class_init (GstAMAScalerPadClass * klass)
{
}

typedef struct
{
  GstXma *xma;
  GstXrm *xrm;
  XmaScalerSession *scaler_sessions[2];
  guint session_send_idx;
  guint session_recv_idx;

  GstXMAFrameAllocator *allocator;

  GstPad *sinkpad;
  GSList *srcpads;
  GHashTable *pad_indexes;
  GList *srcpads_sorted;

  GstVideoInfo in_vinfo;
  gboolean mix_framerate;

  gint32 enable_pipeline;
  GSList *ts_queue;

  gboolean crop;
  guint32 crop_left;
  guint32 crop_top;
  guint32 crop_width;
  guint32 crop_height;
} GstAMAScalerPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (GstAMAScaler, gst_ama_scaler, GST_TYPE_ELEMENT);
GST_ELEMENT_REGISTER_DEFINE (ama_scaler, "ama_scaler", GST_RANK_NONE, GST_TYPE_AMA_SCALER);

static XmaScalerSession *
gst_ama_scaler_create_session (GstAMAScaler * self,
    XmaScalerSession *mix_rate_session)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  XmaScalerSession *result = NULL;
  XmaScalerProperties scaler_props = {0};
  GList *it;
  guint i;

  scaler_props.hwscaler_type = XMA_ABR_SCALER_TYPE;
  scaler_props.handle = priv->xma->xma;
  scaler_props.input.format = XMA_VPE_FMT_TYPE;
  scaler_props.input.sw_format =
      gst_xma_get_format (GST_VIDEO_INFO_FORMAT (&priv->in_vinfo),
          &scaler_props.input.flags);
  scaler_props.input.width = GST_VIDEO_INFO_WIDTH (&priv->in_vinfo);
  scaler_props.input.height = GST_VIDEO_INFO_HEIGHT (&priv->in_vinfo);

  scaler_props.input.framerate.numerator = GST_VIDEO_INFO_FPS_N (&priv->in_vinfo);
  scaler_props.input.framerate.denominator = GST_VIDEO_INFO_FPS_D (&priv->in_vinfo);
  if (priv->mix_framerate && !mix_rate_session) {
    gst_util_fraction_multiply (
        scaler_props.input.framerate.numerator,
        scaler_props.input.framerate.denominator,
        1,
        2,
        &scaler_props.input.framerate.numerator,
        &scaler_props.input.framerate.denominator);
  }

  XmaParameter params[6];
  scaler_props.params = params;
  scaler_props.param_cnt = 0;

  if (priv->crop) {
    if (priv->crop_left <= scaler_props.input.width &&
        (priv->crop_left % 2 == 0) &&
        priv->crop_top <= scaler_props.input.height &&
        (priv->crop_top % 2 == 0) &&
        priv->crop_width <= (scaler_props.input.width - priv->crop_left) &&
        (priv->crop_width % 2 == 0) &&
        priv->crop_height <= (scaler_props.input.height - priv->crop_top) &&
        (priv->crop_height % 2 == 0)) {
      params[scaler_props.param_cnt].name = "left";
      params[scaler_props.param_cnt].type = XMA_INT32;
      params[scaler_props.param_cnt].length = sizeof (gint32);
      params[scaler_props.param_cnt].value = &priv->crop_left;
      ++scaler_props.param_cnt;

      params[scaler_props.param_cnt].name = "top";
      params[scaler_props.param_cnt].type = XMA_INT32;
      params[scaler_props.param_cnt].length = sizeof (gint32);
      params[scaler_props.param_cnt].value = &priv->crop_top;
      ++scaler_props.param_cnt;

      params[scaler_props.param_cnt].name = "width";
      params[scaler_props.param_cnt].type = XMA_INT32;
      params[scaler_props.param_cnt].length = sizeof (gint32);
      params[scaler_props.param_cnt].value = &priv->crop_width;
      ++scaler_props.param_cnt;

      params[scaler_props.param_cnt].name = "height";
      params[scaler_props.param_cnt].type = XMA_INT32;
      params[scaler_props.param_cnt].length = sizeof (gint32);
      params[scaler_props.param_cnt].value = &priv->crop_height;
      ++scaler_props.param_cnt;
    } else {
      GST_WARNING_OBJECT (self, "Crop coordinates invalid; ignoring.");
    }
  }

  params[scaler_props.param_cnt].name = XMA_SCALER_PARAM_THREADS;
  params[scaler_props.param_cnt].length = sizeof (gint32);
  params[scaler_props.param_cnt].value = &priv->enable_pipeline;
  ++scaler_props.param_cnt;

  if (mix_rate_session) {
    params[scaler_props.param_cnt].name = XMA_SCALER_PARAM_MIX_RATE;
    params[scaler_props.param_cnt].length = sizeof (void*);
    params[scaler_props.param_cnt].value = mix_rate_session;
    ++scaler_props.param_cnt;
  }

  scaler_props.num_outputs = 0;
  for (it = priv->srcpads_sorted; it; it = g_list_next (it)) {
    GstAMAScalerPad *srcpad = it->data;

    if (mix_rate_session && srcpad->half_framerate) {
      continue;
    }

    scaler_props.output[scaler_props.num_outputs].format = XMA_VPE_FMT_TYPE;
    scaler_props.output[scaler_props.num_outputs].sw_format =
        gst_xma_get_format (GST_VIDEO_INFO_FORMAT (&srcpad->vinfo),
            &scaler_props.output[scaler_props.num_outputs].flags);
    scaler_props.output[scaler_props.num_outputs].width =
        GST_VIDEO_INFO_WIDTH (&srcpad->vinfo);
    scaler_props.output[scaler_props.num_outputs].height =
        GST_VIDEO_INFO_HEIGHT (&srcpad->vinfo);

    /* Real frame rates to be passed to XRM. */
    scaler_props.output[scaler_props.num_outputs].framerate.numerator =
        GST_VIDEO_INFO_FPS_N (&srcpad->vinfo);
    scaler_props.output[scaler_props.num_outputs].framerate.denominator =
        GST_VIDEO_INFO_FPS_D (&srcpad->vinfo);

    ++scaler_props.num_outputs;
  }

  if (!mix_rate_session) {
    priv->xrm = gst_xrm_reserve_scaler_resources (priv->xma->device,
        &scaler_props);

    if (!priv->xrm) {
      GST_ELEMENT_ERROR (self, RESOURCE, FAILED,
          ("Unable to reserve XRM resources for scaling"), (NULL));
      goto out;
    }
  }

  /* Scaler session input and output frame rates must match. */
  for (i = 0; i != scaler_props.num_outputs; ++i) {
    scaler_props.output[i].framerate.numerator =
        scaler_props.input.framerate.numerator;
    scaler_props.output[i].framerate.denominator =
        scaler_props.input.framerate.denominator;
  }

  result = xma_scaler_session_create (&scaler_props);
  if (!result) {
    GST_ELEMENT_ERROR (self, LIBRARY, INIT,
        ("Could not create scaler session"), (NULL));
  }

out:
  return result;
}

static gboolean
gst_ama_scaler_update_session (GstAMAScaler * self, gboolean create)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  gboolean result = FALSE;
  GstPad *sinkpad = NULL;
  GstCaps *sinkcaps = NULL;
  guint i;

  for (i = 0; i != G_N_ELEMENTS (priv->scaler_sessions); ++i) {
    if (!priv->scaler_sessions[i]) {
      continue;
    }

    if (xma_scaler_session_destroy (priv->scaler_sessions[i]) != XMA_SUCCESS) {
      GST_ERROR_OBJECT (self, "Failed to destroy scaler session");
      return FALSE;
    }
    priv->scaler_sessions[i] = NULL;
  }

  g_clear_pointer (&priv->xrm, gst_xrm_free);

  if (!create) {
    return TRUE;
  }

  priv->scaler_sessions[0] = gst_ama_scaler_create_session (self, NULL);
  if (!priv->scaler_sessions[0]) {
    goto out;
  }

  if (priv->mix_framerate) {
    priv->scaler_sessions[1] = gst_ama_scaler_create_session (self,
        priv->scaler_sessions[0]);

    if (!priv->scaler_sessions[1]) {
      goto out;
    }
  }

  priv->session_send_idx = 0;
  priv->session_recv_idx = 0;

  result = TRUE;
out:
  gst_clear_object (&sinkpad);
  gst_clear_caps (&sinkcaps);

  return result;
}

static gboolean
gst_ama_scaler_start (GstAMAScaler * self)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  priv->allocator = GST_XMA_FRAME_ALLOCATOR (gst_xma_frame_allocator_new ());

  return TRUE;
}

static gboolean
gst_ama_scaler_stop (GstAMAScaler * self)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  g_clear_list (&priv->srcpads_sorted, NULL);
  g_clear_object (&priv->xma);
  g_clear_object (&priv->allocator);
  g_slist_free_full (g_steal_pointer(&priv->ts_queue), (GDestroyNotify)gst_buffer_unref);

  return TRUE;
}

static GstStateChangeReturn
gst_ama_scaler_change_state (GstElement * element, GstStateChange transition)
{
  GstAMAScaler *self = GST_AMA_SCALER (element);
  GstStateChangeReturn ret;

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      if (!gst_ama_scaler_start (self)) {
        return GST_STATE_CHANGE_FAILURE;
      }
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (gst_ama_scaler_parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      if (!gst_ama_scaler_stop (self)) {
        return GST_STATE_CHANGE_FAILURE;
      }
      break;
    default:
      break;
  }

  return ret;
}

static GstPad *
gst_ama_scaler_request_new_pad (GstElement * element,
    GstPadTemplate * templ, const gchar * name_templ, const GstCaps * caps)
{
  GstAMAScaler *self = GST_AMA_SCALER (element);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GstPad *srcpad = NULL;
  gchar *name = NULL;
  guint index = 0;

  GST_DEBUG_OBJECT (self, "Requesting pad");

  GST_OBJECT_LOCK (self);

  if (GST_STATE (self) > GST_STATE_NULL) {
    GST_ERROR_OBJECT (self, "Adding pads is supported only when state is NULL");
    goto out;
  }

  if (element->numsrcpads == MAX_SCALER_OUTPUTS) {
    GST_ERROR_OBJECT (self, "Reached maximum supported outputs");
    goto out;
  }

  if (name_templ && sscanf (name_templ, "src_%u", &index) == 1) {
    GST_LOG_OBJECT (self, "name: %s (index %d)", name_templ, index);
    if (g_hash_table_contains (priv->pad_indexes, GUINT_TO_POINTER (index))) {
      GST_ERROR_OBJECT (self, "Pad name %s is not unique", name_templ);
      goto out;
    }
  } else if (name_templ) {
    GST_ERROR_OBJECT (self, "Incorrect pad name: %s", name_templ);
    goto out;
  }

  g_hash_table_insert (priv->pad_indexes, GUINT_TO_POINTER (index), NULL);

  name = g_strdup_printf ("src_%u", index);
  srcpad = GST_PAD (g_object_new (GST_TYPE_AMA_ABRSCALER_PAD, "name", name,
      "direction", templ->direction, "template", templ, NULL));
  g_clear_pointer (&name, g_free);

  priv->srcpads = g_slist_append (priv->srcpads, srcpad);

out:
  GST_OBJECT_UNLOCK (self);

  gst_element_add_pad (element, srcpad);

  return srcpad;
}

static void
gst_ama_scaler_release_pad (GstElement * element, GstPad * pad)
{
  GstAMAScaler *self = GST_AMA_SCALER (element);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GstAMAScalerPad *srcpad = GST_AMA_SCALER_PAD (pad);

  GST_OBJECT_LOCK (self);

  if (GST_STATE (self) > GST_STATE_NULL) {
    GST_ERROR_OBJECT (self, "Releasing pads is supported only when state is NULL");
    goto out;
  }

  GST_DEBUG_OBJECT (self, "Releasing pad %s", GST_PAD_NAME (pad));

  GST_OBJECT_UNLOCK (self);

  gst_object_ref (pad);
  gst_element_remove_pad (element, pad);

  GST_OBJECT_LOCK (self);

  g_hash_table_remove (priv->pad_indexes, GUINT_TO_POINTER (srcpad->index));
  priv->srcpads = g_slist_remove (priv->srcpads, srcpad);
  gst_object_unref (pad);

out:
  GST_OBJECT_UNLOCK (self);
}

static GstCaps *
gst_ama_scaler_get_peer_caps (GstAMAScaler * self, GstPad * srcpad)
{
  GstPad *srcpeer;
  GstCaps *peercaps;

  srcpeer = gst_pad_get_peer (srcpad);
  peercaps = gst_pad_query_caps (srcpeer, NULL);
  gst_clear_object (&srcpeer);

  return peercaps;
}

static gboolean
copy_framerate_to_caps (GstCapsFeatures * features,
    GstStructure * structure, gpointer user_data)
{
  if (gst_structure_has_field (structure, "framerate")) {
    GstCaps *caps = user_data;
    GstStructure *s = gst_structure_new_id_empty (structure->name);

    gst_structure_set_value (s, "framerate",
        gst_structure_get_value (structure, "framerate"));

    gst_caps_append_structure_full(caps, s, gst_caps_features_copy (features));
  }

  return TRUE;
}

static void
set_half_framerate (GstCaps * caps)
{
  GstStructure *s;
  gint fps_n, fps_d;

  g_assert (gst_caps_get_size (caps) == 1);

  s = gst_caps_get_structure (caps, 0);

  gst_structure_get_fraction (s, "framerate", &fps_n, &fps_d);
  gst_util_fraction_multiply (fps_n, fps_d, 1, 2, &fps_n, &fps_d);
  gst_structure_set (s, "framerate", GST_TYPE_FRACTION, fps_n, fps_d, NULL);
}

static GstCaps *
gst_ama_scaler_get_fixated_caps (GstAMAScaler * self, GstPad * srcpad,
    const GstCaps * incaps)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GstCaps *peercaps;
  GstCaps *filter;
  GstCaps *result;
  GstStructure *ins, *outs;

  peercaps = gst_ama_scaler_get_peer_caps (self, srcpad);

  /* This is basically choosing common resolution between sink and src. */
  filter = gst_caps_copy (incaps);
  gst_structure_remove_fields (gst_caps_get_structure (filter, 0),
      "format", "width", "height", NULL);

  if (gst_caps_can_intersect (peercaps, filter)) {
    GST_AMA_SCALER_PAD (srcpad)->half_framerate = FALSE;
  } else {
    /* Downstream doesn't accept full framerate. Try half framerate. */
    set_half_framerate (filter);
    GST_AMA_SCALER_PAD (srcpad)->half_framerate = TRUE;
    priv->mix_framerate = TRUE;
  }

  result = gst_caps_intersect (peercaps, filter);
  result = gst_caps_truncate (result);
  gst_clear_caps (&filter);

  GST_DEBUG_OBJECT (self, "Fixating src caps %" GST_PTR_FORMAT
      " based on sink caps %" GST_PTR_FORMAT, result, incaps);

  ins = gst_caps_get_structure (incaps, 0);
  outs = gst_caps_get_structure (result, 0);

  {
    const gchar *in_format;

    in_format = gst_structure_get_string (ins, "format");
    if (in_format) {
      gst_structure_fixate_field_string (outs, "format", in_format);
    }
  }

  {
    gint in_width;

    if (gst_structure_get_int (ins, "width", &in_width)) {
      if (gst_structure_has_field (outs, "width")) {
        gst_structure_fixate_field_nearest_int (outs, "width", in_width);
      } else {
        gst_structure_set (outs, "width", G_TYPE_INT, in_width, NULL);
      }
    }
  }

  {
    gint in_height;

    if (gst_structure_get_int (ins, "height", &in_height)) {
      if (gst_structure_has_field (outs, "height")) {
        gst_structure_fixate_field_nearest_int (outs, "height", in_height);
      } else {
        gst_structure_set (outs, "height", G_TYPE_INT, in_height, NULL);
      }
    }
  }

  gst_clear_caps (&peercaps);

  /* Fixate the remaining fields. */
  return gst_caps_fixate (result);
}

static gint
cmp_scalerpads (const GstAMAScalerPad * a, const GstAMAScalerPad * b)
{
  /* XMA scaler session wants XmaScalerProperties.output in descending order
   * of resolution and half fps outputs must appear after full fps ones. */

  if (a->half_framerate && !b->half_framerate) {
    return 1;
  } else if (!a->half_framerate && b->half_framerate) {
    return -1;
  } else {
    gint r;

    r = GST_VIDEO_INFO_WIDTH (&b->vinfo) - GST_VIDEO_INFO_WIDTH (&a->vinfo);
    if (r == 0) {
      r = GST_VIDEO_INFO_HEIGHT (&b->vinfo) - GST_VIDEO_INFO_HEIGHT (&a->vinfo);
    }

    return r;
  }
}

static gboolean
gst_ama_scaler_sink_setcaps (GstAMAScaler * self, GstPad * sinkpad,
    GstCaps * incaps)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  gint32 device;
  GSList *it;

  g_return_val_if_fail (gst_caps_is_fixed (incaps), FALSE);

  GST_DEBUG_OBJECT (self, "Have new sink caps %p %" GST_PTR_FORMAT, incaps,
      incaps);

  priv->mix_framerate = FALSE;
  g_clear_list (&priv->srcpads_sorted, NULL);

  if (!gst_video_info_from_caps (&priv->in_vinfo, incaps)) {
    GST_ERROR_OBJECT (self,
        "Error parsing video info from caps %" GST_PTR_FORMAT, incaps);
    return FALSE;
  }

  device = gst_ama_device_get_from_caps (incaps);
  if (device == GST_XMA_DEVICE_UNSPECIFIED) {
    g_clear_object (&priv->xma);
    GST_ELEMENT_ERROR (self, CORE, CAPS,
        ("Invalid or missing 'ama-device' in sink caps"), (NULL));
    return FALSE;
  }

  if (!priv->xma || priv->xma->device != device) {
    g_clear_object (&priv->xma);

    priv->xma = gst_xma_get_for_device (device);
    if (!priv->xma) {
      GST_ELEMENT_ERROR (self, LIBRARY, INIT,
          ("Failed to initialize XMA library"), (NULL));
      return FALSE;
    }
  }

  for (it = priv->srcpads; it; it = g_slist_next (it)) {
    GstPad *srcpad = it->data;
    GstAMAScalerPad *scalerpad = GST_AMA_SCALER_PAD (srcpad);
    GstCaps *outcaps;
    gboolean ret;

    outcaps = gst_ama_scaler_get_fixated_caps (self, srcpad, incaps);
    if (!outcaps) {
      return FALSE;
    }

    GST_DEBUG_OBJECT (self, "Setting caps %" GST_PTR_FORMAT " on %s", outcaps,
        GST_PAD_NAME (srcpad));

    ret = gst_pad_set_caps (srcpad, outcaps);

    if (!gst_video_info_from_caps (&scalerpad->vinfo, outcaps)) {
      GST_ERROR_OBJECT (self,
          "Error parsing video info from caps %" GST_PTR_FORMAT, outcaps);
      return FALSE;
    }

    gst_clear_caps (&outcaps);
    if (!ret) {
      return FALSE;
    }

    priv->srcpads_sorted = g_list_insert_sorted (priv->srcpads_sorted,
        scalerpad, (GCompareFunc) cmp_scalerpads);
  }

  return TRUE;
}

static GstFlowReturn
gst_ama_scaler_receive_xma_frames (GstAMAScaler * self)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  XmaScalerSession *current_session = priv->scaler_sessions[priv->session_recv_idx];
  XmaFrame *frames[MAX_SCALER_OUTPUTS] = {0};
  GstBuffer *inbuf = NULL;
  int32_t xmares;
  GList *it;
  guint i;

  do {
    xmares = xma_scaler_session_recv_frame_list (current_session, frames);
    if (xmares == XMA_TRY_AGAIN) {
      if (priv->enable_pipeline) {
        return GST_FLOW_OK;
      }
      g_usleep (100);
    }
  } while (xmares == XMA_TRY_AGAIN);

  if (xmares == XMA_EOS) {
    return GST_FLOW_EOS;
  } else if (xmares != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, LIBRARY, FAILED,
        ("Could not receive scaled frames: %d", xmares), (NULL));
    return GST_FLOW_ERROR;
  }

  if (priv->ts_queue) {
    inbuf = priv->ts_queue->data;
  }
  for (it = priv->srcpads_sorted, i = 0; it; it = g_list_next (it)) {
    GstAMAScalerPad *srcpad = it->data;
    GstMemory *mem;
    GstBuffer *buf;
    GstFlowReturn ret;

    if (priv->session_recv_idx == 1 && srcpad->half_framerate) {
      break;
    }

    mem = gst_xma_frame_allocator_wrap_frame (priv->allocator,
        GST_VIDEO_INFO_SIZE(&srcpad->vinfo), priv->xma, frames[i++]);

    buf = gst_buffer_new ();
    gst_buffer_append_memory (buf, mem);

    if (inbuf) {
      GST_BUFFER_PTS (buf) = GST_BUFFER_PTS (inbuf);
      GST_BUFFER_DTS (buf) = GST_BUFFER_DTS (inbuf);
      GST_BUFFER_DURATION (buf) = GST_BUFFER_DURATION (inbuf);
    }

    ret = gst_pad_push (GST_PAD (srcpad), buf);
    if (ret != GST_FLOW_OK) {
      GST_ELEMENT_ERROR (self, CORE, PAD, ("Could not push frame"), (NULL));
      return ret;
    }
  }
  if (inbuf) {
    priv->ts_queue = g_slist_remove(priv->ts_queue, inbuf);
    gst_buffer_unref(inbuf);
  }

  if (priv->mix_framerate) {
    priv->session_recv_idx = (priv->session_recv_idx + 1) % 2;
  }

  return GST_FLOW_OK;
}

static gboolean
gst_ama_scaler_sink_event (GstPad * pad, GstObject * parent, GstEvent * event)
{
  GstAMAScaler *self = GST_AMA_SCALER (parent);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  gboolean ret = TRUE;

  GST_DEBUG_OBJECT (pad, "Received event '%s' %p %" GST_PTR_FORMAT,
      gst_event_type_get_name (GST_EVENT_TYPE (event)), event, event);

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_CAPS:{
      GstCaps *caps;

      gst_event_parse_caps (event, &caps);
      ret = gst_ama_scaler_sink_setcaps (self, priv->sinkpad, caps);
      gst_clear_event (&event);
      break;
    }
    case GST_EVENT_EOS:
      xma_scaler_session_send_frame (priv->scaler_sessions[priv->session_send_idx],
          NULL);
      while ((ret = gst_ama_scaler_receive_xma_frames (self)) == GST_FLOW_OK);
      if (ret == GST_FLOW_EOS) {
        ret = gst_pad_event_default (pad, parent, event);
      }
      break;
    default:
      ret = gst_pad_event_default (pad, parent, event);
      break;
  }

  return ret;
}

static void
append_multiplied_fraction (GValue * list, const GValue * fraction)
{
  GValue new_fraction = G_VALUE_INIT;

  g_value_init (&new_fraction, GST_TYPE_FRACTION);

  gst_value_set_fraction (&new_fraction,
      2 * gst_value_get_fraction_numerator (fraction),
      gst_value_get_fraction_denominator (fraction));

  gst_value_list_append_value (list, &new_fraction);

  g_value_unset (&new_fraction);
}

static void
add_double_framerates_to_caps (GstCaps * caps)
{
  const GValue *val;

  g_assert (gst_caps_get_size (caps) == 1);

  val = gst_structure_get_value (gst_caps_get_structure (caps, 0), "framerate");

  if (GST_VALUE_HOLDS_LIST (val) || GST_VALUE_HOLDS_FRACTION (val)) {
    GValue new_list = G_VALUE_INIT;
    gint i;

    if (GST_VALUE_HOLDS_LIST (val)) {
      g_value_copy (val, &new_list);

      for (i = 0; i != gst_value_list_get_size (val); ++i) {
        const GValue *v = gst_value_list_get_value (val, i);

        if (GST_VALUE_HOLDS_FRACTION (v)) {
          append_multiplied_fraction (&new_list, v);
        }
      }
    } else if (GST_VALUE_HOLDS_FRACTION (val)) {
      gst_value_list_init (&new_list, 2);
      gst_value_list_append_value (&new_list, val);

      append_multiplied_fraction (&new_list, val);
    }

    gst_caps_set_value (caps, "framerate", &new_list);

    g_value_unset (&new_list);
  }
}

static GValue *
gst_ama_scaler_collect_src_framerates (GstAMAScaler * self)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GSList *it;
  GstCaps *framerate_caps = gst_caps_new_empty();
  GstCaps *templatecaps = gst_pad_get_pad_template_caps (priv->sinkpad);
  GValue *result = NULL;

  copy_framerate_to_caps (gst_caps_get_features (templatecaps, 0),
      gst_caps_get_structure (templatecaps, 0), framerate_caps);

  gst_clear_caps (&templatecaps);

  for (it = priv->srcpads;
      !gst_caps_is_empty (framerate_caps) && it; it = g_slist_next (it)) {
    GstPad *pad = it->data;
    GstCaps *caps = gst_ama_scaler_get_peer_caps (self, pad);
    GstCaps *pad_framerate_caps = gst_caps_new_empty ();

    gst_caps_foreach (caps, copy_framerate_to_caps, pad_framerate_caps);

    if (gst_caps_is_empty (pad_framerate_caps)) {
      GST_LOG_OBJECT (self, "%s supports any framerate", GST_PAD_NAME (pad));
    } else {
      GstCaps *tmp;

      pad_framerate_caps = gst_caps_simplify (pad_framerate_caps);

      GST_LOG_OBJECT (self, "%s supports framerates %" GST_PTR_FORMAT,
          GST_PAD_NAME (pad), pad_framerate_caps);

      add_double_framerates_to_caps (pad_framerate_caps);

      tmp = gst_caps_intersect (framerate_caps, pad_framerate_caps);
      gst_clear_caps (&framerate_caps);
      framerate_caps = tmp;
    }

    gst_clear_caps (&pad_framerate_caps);
    gst_clear_caps (&caps);
  }

  framerate_caps = gst_caps_simplify (framerate_caps);

  GST_DEBUG_OBJECT (self, "Src pads support framerates: %" GST_PTR_FORMAT,
      framerate_caps);

  if (!gst_caps_is_empty (framerate_caps)) {
    result = g_new0 (GValue, 1);
    gst_value_init_and_copy (result,
        gst_structure_get_value (gst_caps_get_structure (framerate_caps, 0),
            "framerate"));
  }

  gst_clear_caps (&framerate_caps);

  return result;
}

typedef struct
{
  GValue *framerates;
  GstCaps *output_caps;
} CapsCtx;

static gboolean
append_caps_with_framerates (GstCapsFeatures * features,
    GstStructure * structure, gpointer user_data)
{
  CapsCtx *ctx = user_data;
  GstStructure *s = gst_structure_copy (structure);

  gst_structure_set_value (s, "framerate", ctx->framerates);

  gst_caps_append_structure_full (ctx->output_caps, s,
      gst_caps_features_copy (features));

  return TRUE;
}

static GstCaps *
gst_ama_scaler_query_caps (GstAMAScaler * self, GstCaps * filter)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GstCaps *result = gst_caps_new_empty();
  GstCaps *templatecaps;
  CapsCtx ctx = { gst_ama_scaler_collect_src_framerates (self), result };
  GSList *it;

  if (!ctx.framerates) {
    GST_ELEMENT_ERROR (self, CORE, NEGOTIATION,
        ("Couldn't negotiate common framerate on src pads."), (NULL));
    goto out;
  }

  /* Sink can only negotiate a framerate ALL src pads can accept. */

  for (it = priv->srcpads; it; it = g_slist_next (it)) {
    GstPad *srcpad = it->data;
    GstCaps *peercaps;

    peercaps = gst_ama_scaler_get_peer_caps (self, srcpad);
    gst_caps_foreach (peercaps, append_caps_with_framerates, &ctx);
    gst_clear_caps (&peercaps);
  }

  templatecaps = gst_pad_get_pad_template_caps (priv->sinkpad);
  gst_caps_foreach (templatecaps, append_caps_with_framerates, &ctx);
  gst_clear_caps (&templatecaps);

  g_value_unset (ctx.framerates);
  g_free (ctx.framerates);

  if (filter) {
    GstCaps *filter_caps;

    GST_LOG_OBJECT (self, "Intersecting %" GST_PTR_FORMAT " with %" GST_PTR_FORMAT,
        result, filter);
    filter_caps = gst_caps_intersect (result, filter);
    gst_clear_caps (&result);
    result = filter_caps;
  }

out:
  /* The returned caps in order of preference:
   * - peer caps, which may have explicit resolutions and/or framerates
   * - sink template caps; no explicit resolution, may have explicit framerate */
  return result;
}

static gboolean
gst_ama_scaler_sink_query (GstPad * pad, GstObject * parent, GstQuery * query)
{
  GstAMAScaler *self = GST_AMA_SCALER (parent);

  gboolean ret = TRUE;

  switch (GST_QUERY_TYPE (query)) {
    case GST_QUERY_CAPS:{
      GstCaps *filter;
      GstCaps *caps;

      gst_query_parse_caps (query, &filter);
      caps = gst_ama_scaler_query_caps (self, filter);

      if (caps) {
        GST_DEBUG_OBJECT (self, "Returning %s caps %" GST_PTR_FORMAT,
            GST_PAD_NAME (pad), caps);
        gst_query_set_caps_result (query, caps);
        gst_clear_caps (&caps);
      } else {
        ret = FALSE;
      }
      break;
    }
    default:
      ret = gst_pad_query_default (pad, parent, query);
      break;
  }

  return ret;
}

static GstFlowReturn
gst_ama_scaler_chain (GstPad * pad, GstObject * parent, GstBuffer * inbuf)
{
  GstAMAScaler *self = GST_AMA_SCALER (parent);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  GstMemory *mem;
  int32_t xmares;
  GstFlowReturn ret = GST_FLOW_ERROR;

  if (!priv->scaler_sessions[0] && !gst_ama_scaler_update_session (self, TRUE)) {
    goto done;
  }

  mem = gst_buffer_peek_memory (inbuf, 0);
  if (!gst_is_xma_frame (mem)) {
    GST_ELEMENT_ERROR (self, CORE, NOT_IMPLEMENTED,
        ("invalid input video buffer"), (NULL));
    goto done;
  }

  priv->ts_queue = g_slist_append (priv->ts_queue, gst_buffer_ref (inbuf));
  xma_frame_inc_ref (GST_XMA_FRAME_CAST (mem)->frame);
  while (TRUE) {
    xmares = xma_scaler_session_send_frame (priv->scaler_sessions[priv->session_send_idx],
        GST_XMA_FRAME_CAST (mem)->frame);
    if (priv->enable_pipeline && xmares == XMA_TRY_AGAIN) {
      if (gst_ama_scaler_receive_xma_frames (self) != GST_FLOW_OK) {
        ret = GST_FLOW_ERROR;
        goto done;
      }
    } else {
      break;
    }
  }

  if (priv->mix_framerate) {
    priv->session_send_idx = (priv->session_send_idx + 1) % 2;
  }

  if (xmares != XMA_SUCCESS && xmares != XMA_SEND_MORE_DATA) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL), ("failed to send frame"));
    goto done;
  }

  if (xmares == XMA_SEND_MORE_DATA) {
    ret = GST_FLOW_OK;
  } else {
    ret = gst_ama_scaler_receive_xma_frames (self);
  }

done:
  gst_buffer_unref (inbuf);
  return ret;
}

static void
gst_ama_scaler_init (GstAMAScaler * self)
{
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  priv->sinkpad = gst_pad_new_from_static_template (&sink_template, "sink");
  gst_pad_set_event_function (priv->sinkpad,
      GST_DEBUG_FUNCPTR (gst_ama_scaler_sink_event));
  gst_pad_set_query_function (priv->sinkpad,
        GST_DEBUG_FUNCPTR (gst_ama_scaler_sink_query));
  gst_pad_set_chain_function (priv->sinkpad,
      GST_DEBUG_FUNCPTR (gst_ama_scaler_chain));
  gst_element_add_pad (GST_ELEMENT (self), priv->sinkpad);

  priv->pad_indexes = g_hash_table_new (NULL, NULL);
}

static void
gst_ama_scaler_set_property (GObject * object, guint prop_id, const GValue * value,
    GParamSpec * pspec)
{
  GstAMAScaler *self = GST_AMA_SCALER (object);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  switch (prop_id) {
    case PROP_ENABLE_PIPELINE:
      priv->enable_pipeline = g_value_get_boolean (value) ? 1 : 0;
      break;
    case PROP_CROP:
      priv->crop = g_value_get_boolean (value);
      break;
    case PROP_CROP_LEFT:
      priv->crop_left = g_value_get_uint (value);
      break;
    case PROP_CROP_TOP:
      priv->crop_top = g_value_get_uint (value);
      break;
    case PROP_CROP_WIDTH:
      priv->crop_width = g_value_get_uint (value);
      break;
    case PROP_CROP_HEIGHT:
      priv->crop_height = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_scaler_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAMAScaler *self = GST_AMA_SCALER (object);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      if (priv->xma) {
        g_value_set_int (value, priv->xma->device);
      } else {
        g_value_set_int (value, GST_XMA_DEVICE_UNSPECIFIED);
      }
      break;
    case PROP_ENABLE_PIPELINE:
      g_value_set_boolean (value, priv->enable_pipeline);
      break;
    case PROP_CROP:
      g_value_set_boolean (value, priv->crop);
      break;
    case PROP_CROP_LEFT:
      g_value_set_uint (value, priv->crop_left);
      break;
    case PROP_CROP_TOP:
      g_value_set_uint (value, priv->crop_top);
      break;
    case PROP_CROP_WIDTH:
      g_value_set_uint (value, priv->crop_width);
      break;
    case PROP_CROP_HEIGHT:
      g_value_set_uint (value, priv->crop_height);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_scaler_finalize (GObject * object)
{
  GstAMAScaler *self = GST_AMA_SCALER (object);
  GstAMAScalerPrivate *priv = gst_ama_scaler_get_instance_private (self);

  g_hash_table_unref (priv->pad_indexes);
  gst_ama_scaler_update_session (self, FALSE);

  G_OBJECT_CLASS (gst_ama_scaler_parent_class)->finalize (object);
}

static void
gst_ama_scaler_class_init (GstAMAScalerClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gobject_class->set_property = gst_ama_scaler_set_property;
  gobject_class->get_property = gst_ama_scaler_get_property;
  gobject_class->finalize = gst_ama_scaler_finalize;

  element_class->change_state = gst_ama_scaler_change_state;
  element_class->request_new_pad = gst_ama_scaler_request_new_pad;
  element_class->release_pad = gst_ama_scaler_release_pad;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_int ("device", "XMA device",
          "Numeric ID of an XMA device (-1 = unspecified)",
          -1, G_MAXINT32, GST_XMA_DEVICE_UNSPECIFIED,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, PROP_ENABLE_PIPELINE,
      g_param_spec_boolean ("enable-pipeline", "Enable pipeline",
          "Enable pipelining for XABR scaler",
          TRUE,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CROP,
      g_param_spec_boolean ("crop", "Crop",
          "Crop the input frame before scaling.",
          FALSE,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CROP_LEFT,
      g_param_spec_uint ("crop-left", "Crop left",
          "Left position of crop area. Must be < input width and a multiple of 2",
          0, G_MAXINT32, 0,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CROP_TOP,
      g_param_spec_uint ("crop-top", "Crop top",
          "Top position of crop area. Must be < input height and a multiple of 2",
          0, G_MAXINT32, 0,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CROP_WIDTH,
      g_param_spec_uint ("crop-width", "Crop width",
          "Width of crop area. Must be < input width - crop-left and a multiple of 2",
          0, G_MAXINT32, 0,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_CROP_HEIGHT,
      g_param_spec_uint ("crop-height", "Crop height",
          "Height of crop area. Must be < input height - crop-top and a multiple of 2",
          0, G_MAXINT32, 0,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  gst_element_class_add_static_pad_template (element_class, &sink_template);
  gst_element_class_add_static_pad_template (element_class, &src_template);

  gst_element_class_set_static_metadata (element_class,
      "Xilinx AMA ABR scaler", "Filter/Converter/Video/Scaler",
      "Xilinx AMA ABR scaler", "Xilinx Inc., https://www.xilinx.com");

  GST_DEBUG_CATEGORY_INIT (gst_ama_scaler_debug, "ama_scaler", 0,
      "Xilinx AMA Scaler");
}
