/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamaupload.h"
#include "gstamacommon.h"
#include "gstxmaframeallocator.h"

#include <gst/video/video-format.h>
#include <gst/video/video-info.h>

#include <xmaerror.h>
#include <xmafilter.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_upload_debug);
#define GST_CAT_DEFAULT gst_ama_upload_debug

enum
{
  PROP_0,
  PROP_DEVICE,
};

static GstStaticPadTemplate sink_template = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE (GST_AMA_VIDEO_FORMATS))
    );

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_MAKE_WITH_FEATURES
        (GST_CAPS_FEATURE_MEMORY_AMA_MEMORY, GST_AMA_VIDEO_FORMATS))
    );

typedef struct
{
  gint32 device;
  GstXma *xma;
  XmaFilterSession *upload_session;

  GstXMAFrameAllocator *allocator;

  GstVideoInfo info;
} GstAMAUploadPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (GstAMAUpload, gst_ama_upload, GST_TYPE_BASE_TRANSFORM);
GST_ELEMENT_REGISTER_DEFINE (ama_upload, "ama_upload", GST_RANK_NONE, GST_TYPE_AMA_UPLOAD);

static gboolean
gst_ama_upload_start (GstBaseTransform * transform)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (transform);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  priv->xma = gst_xma_get_for_device (priv->device);
  if (!priv->xma) {
    GST_ELEMENT_ERROR (self, LIBRARY, INIT,
        ("Failed to initialize XMA library"), (NULL));
    return FALSE;
  }

  priv->allocator = GST_XMA_FRAME_ALLOCATOR (gst_xma_frame_allocator_new ());

  return TRUE;
}

static GstXMAFrame *
gst_ama_upload_create_host_frame (GstAMAUpload * self)
{
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  XmaFrameProperties frame_props = {0};

  frame_props.format = gst_xma_get_format (priv->info.finfo->format,
      (gint32 *) &frame_props.flags);
  frame_props.sw_format = gst_xma_get_format (priv->info.finfo->format,
      &frame_props.flags);
  frame_props.width = priv->info.width;
  frame_props.height = priv->info.height;

  return GST_XMA_FRAME_CAST (gst_xma_frame_allocator_alloc (priv->allocator,
      GST_VIDEO_INFO_SIZE (&priv->info), priv->xma, &frame_props, FALSE));
}

static gboolean
gst_ama_upload_update_video_info (GstAMAUpload * self, GstVideoInfo * info)
{
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  if (priv->upload_session) {
    if (xma_filter_session_destroy (priv->upload_session) != XMA_SUCCESS) {
      GST_ELEMENT_ERROR (self, LIBRARY, SHUTDOWN,
          ("Failed to destroy upload filter session"), (NULL));
      return FALSE;
    }
    priv->upload_session = NULL;
  }

  if (info) {
    XmaFilterProperties filter_props = {0};

    priv->info = *info;

    filter_props.hwfilter_type = XMA_UPLOAD_FILTER_TYPE;
    filter_props.handle = priv->xma->xma;
    filter_props.input.format = gst_xma_get_format (priv->info.finfo->format,
        &filter_props.input.flags);
    filter_props.input.sw_format = filter_props.input.format;
    filter_props.input.width = GST_VIDEO_INFO_WIDTH (&priv->info);
    filter_props.input.height = GST_VIDEO_INFO_HEIGHT (&priv->info);
    filter_props.input.framerate.numerator    = GST_VIDEO_INFO_FPS_N (&priv->info);
    filter_props.input.framerate.denominator  = GST_VIDEO_INFO_FPS_D (&priv->info);
    filter_props.output.format                = XMA_VPE_FMT_TYPE;
    filter_props.output.sw_format             = filter_props.input.sw_format;
    filter_props.output.width                 = filter_props.input.width;
    filter_props.output.height                = filter_props.input.height;
    filter_props.output.framerate.numerator   = filter_props.input.framerate.numerator;
    filter_props.output.framerate.denominator = filter_props.input.framerate.denominator;

    priv->upload_session = xma_filter_session_create (&filter_props);
    if (!priv->upload_session) {
      GST_ELEMENT_ERROR (self, LIBRARY, SETTINGS,
          ("Failed to create upload filter session"), (NULL));
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
gst_ama_upload_stop (GstBaseTransform * transform)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (transform);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  if (!gst_ama_upload_update_video_info (self, NULL)) {
    return FALSE;
  }

  g_clear_object (&priv->xma);

  return TRUE;
}

static GstCaps *
gst_ama_upload_transform_caps (GstBaseTransform * trans,
    GstPadDirection direction, GstCaps * caps, GstCaps * filter)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (trans);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  GstCaps *tmp = gst_caps_copy (caps);
  GstCaps *result;

  gst_ama_toggle_memory_caps_feature (tmp, direction == GST_PAD_SINK);
  if (direction == GST_PAD_SINK && priv->xma) {
    gst_ama_device_add_to_caps (tmp, priv->xma->device);
  }

  result = GST_BASE_TRANSFORM_CLASS (gst_ama_upload_parent_class)->transform_caps (trans,
      direction, tmp, filter);

  gst_clear_caps (&tmp);

  return result;
}

static gboolean
gst_ama_upload_set_caps (GstBaseTransform * transform, GstCaps *incaps,
    GstCaps *outcaps)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (transform);

  GstVideoInfo in_info, out_info;

  if (!gst_video_info_from_caps (&in_info, incaps)) {
    goto invalid_caps;
  }
  if (!gst_video_info_from_caps (&out_info, outcaps)) {
    goto invalid_caps;
  }

  if (!gst_video_info_is_equal (&in_info, &out_info)) {
    goto format_mismatch;
  }

  return gst_ama_upload_update_video_info (self, &in_info);

invalid_caps:
  GST_ELEMENT_ERROR (self, CORE, CAPS, ("invalid caps"), (NULL));
  return FALSE;

format_mismatch:
  GST_ELEMENT_ERROR (self, CORE, CAPS,
      ("input and output formats do not match"), (NULL));
  return FALSE;
}

static GstFlowReturn
gst_ama_upload_prepare_output_buffer (GstBaseTransform * transform,
    GstBuffer * buffer, GstBuffer ** outbuf)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (transform);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  XmaFrameProperties props = {0};
  GstMemory *mem;
  guint i;

  props.format = XMA_VPE_FMT_TYPE;
  props.sw_format = gst_xma_get_format (priv->info.finfo->format, &props.flags);
  props.width = priv->info.width;
  props.height = priv->info.height;

  for (i = 0; i != GST_VIDEO_INFO_N_PLANES (&priv->info); ++i) {
    props.linesize[i] = GST_VIDEO_INFO_PLANE_STRIDE (&priv->info, i);
  }

  mem = gst_xma_frame_allocator_alloc (priv->allocator,
      GST_VIDEO_INFO_SIZE (&priv->info), priv->xma, &props, TRUE);
  if (!mem) {
    return GST_FLOW_ERROR;
  }

  *outbuf = gst_buffer_new ();
  gst_buffer_append_memory (*outbuf, mem);

  GST_BASE_TRANSFORM_GET_CLASS (transform)->copy_metadata (transform, buffer,
      *outbuf);

  return GST_FLOW_OK;
}

static GstFlowReturn
gst_ama_upload_transform (GstBaseTransform * transform, GstBuffer * inbuf,
    GstBuffer * outbuf)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (transform);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  GstFlowReturn result = GST_FLOW_ERROR;
  GstVideoFrame in_frame;
  GstXMAFrame *host_frame = NULL;
  GstMemory *mem;
  XmaFrame *out_frame;
  guint i;

  if (!gst_video_frame_map (&in_frame, &priv->info, inbuf, GST_MAP_READ)) {
    GST_ELEMENT_ERROR (self, CORE, NOT_IMPLEMENTED, (NULL),
        ("invalid video buffer received"));
    goto out;
  }

  mem = gst_buffer_peek_memory (outbuf, 0);
  if (!gst_is_xma_frame (mem)) {
    GST_ELEMENT_ERROR (self, CORE, NOT_IMPLEMENTED, (NULL),
        ("invalid output video buffer"));
    goto out;
  }

  host_frame = gst_ama_upload_create_host_frame (self);
  if (!host_frame) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL),
        ("Failed to allocate XmaFrame"));
    goto out;
  }

  for (i = 0; i < GST_VIDEO_FRAME_N_PLANES (&in_frame); ++i) {
    guint j;
    const guint input_stride = GST_VIDEO_FRAME_PLANE_STRIDE (&in_frame, i);
    const guint output_stride = host_frame->frame->frame_props.linesize[i];
    const guint plane_height =
        GST_VIDEO_FORMAT_INFO_SCALE_HEIGHT (in_frame.info.finfo, i,
            GST_VIDEO_FRAME_HEIGHT (&in_frame));
    const guint line_length = MIN(input_stride, output_stride);

    for (j = 0; j < plane_height; ++j) {
      memcpy (host_frame->frame->data[i].host + (j * output_stride),
          GST_VIDEO_FRAME_PLANE_DATA (&in_frame, i) + (j * input_stride),
          line_length);
    }
  }

  /* XMA library takes ownership of the host-side XmaFrame. */
  if (xma_filter_session_send_frame (priv->upload_session,
      g_steal_pointer (&host_frame->frame)) != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL),
        ("failed to upload frame to device"));
    goto out;
  }

  out_frame = GST_XMA_FRAME_CAST (mem)->frame;

  if (xma_filter_session_recv_frame (priv->upload_session,out_frame) != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL),
        ("failed to upload frame to device"));
    goto out;
  }

  result = GST_FLOW_OK;

out:
  if (host_frame) {
    gst_memory_unref (GST_MEMORY_CAST (host_frame));
  }
  gst_video_frame_unmap (&in_frame);

  return result;
}

static void
gst_ama_upload_init (GstAMAUpload * upload)
{
}

static void
gst_ama_upload_set_property (GObject * object, guint prop_id, const GValue * value,
    GParamSpec * pspec)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (object);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      priv->device = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_upload_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (object);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      g_value_set_int (value, priv->device);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_upload_dispose (GObject * object)
{
  GstAMAUpload *self = GST_AMA_UPLOAD (object);
  GstAMAUploadPrivate *priv = gst_ama_upload_get_instance_private (self);

  gst_clear_object (&priv->xma);
  gst_clear_object (&priv->allocator);

  G_OBJECT_CLASS (gst_ama_upload_parent_class)->dispose (object);
}


static void
gst_ama_upload_class_init (GstAMAUploadClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstBaseTransformClass *transform_class = GST_BASE_TRANSFORM_CLASS (klass);

  gobject_class->set_property = gst_ama_upload_set_property;
  gobject_class->get_property = gst_ama_upload_get_property;
  gobject_class->dispose = gst_ama_upload_dispose;

  transform_class->start = gst_ama_upload_start;
  transform_class->stop = gst_ama_upload_stop;
  transform_class->transform_caps = gst_ama_upload_transform_caps;
  transform_class->set_caps = gst_ama_upload_set_caps;
  transform_class->prepare_output_buffer = gst_ama_upload_prepare_output_buffer;
  transform_class->transform = gst_ama_upload_transform;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_int ("device", "XMA device",
          "Numeric ID of an XMA device (-1 = auto)",
          -1, G_MAXINT32, GST_XMA_DEVICE_AUTO,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  gst_element_class_add_static_pad_template (element_class, &sink_template);
  gst_element_class_add_static_pad_template (element_class, &src_template);

  gst_element_class_set_static_metadata (element_class,
      "Xilinx AMA uploader", "Filter/Video",
      "Uploads data into Xilinx AMA", "Xilinx Inc., https://www.xilinx.com");

  GST_DEBUG_CATEGORY_INIT (gst_ama_upload_debug, "ama_upload", 0,
      "Xilinx AMA uploader");
}
